from PiQR.conversions import *
import json
# Subdict indexs refer to the following
# Code Word Qty,  Code Words Per Block,  Group1 Blocks,  Block1 Code Word Qty,  Group2 Blocks,  Block2 Code Word Qty

def load_data():
    with open('qr_variables.json', 'r') as f:
        file_data = json.load(f)
    return file_data

qr_variables = load_data()

ecc_dict = qr_variables['ECC']

character_capacities = qr_variables['CharacterCapacities']

alpha_numeric = qr_variables['AlphaNumeric']

mode_indicators = qr_variables['ModeIndicators']

char_fill = qr_variables['CharFill']

def place_data_codewords(qr_size, binary, reserve_pixels):
    start_x, start_y = qr_size - 1, 0
    orientation = 'Up'
    last_direction = ''
    next_direction = 'Left'
    dir_change = False

    total_coords = qr_size * qr_size
    bit_coords = []
    bit_coords.append((start_x, start_y))

    while start_x >= 0:
        if last_direction == 'Left':
            if dir_change:
                next_direction = 'Left'
                start_x -= 1
                if start_x == 6:
                    start_x -= 1
                dir_change = False
            else:
                next_direction = 'Right'
        elif last_direction == 'Right':
            next_direction = 'Left'

        if orientation == 'Up':
            if next_direction == 'Left':
                start_x -= 1
                if (start_x, start_y) in [i[0] for i in reserve_pixels]:
                    pass
                else:
                    if start_x >= 0:
                        bit_coords.append((start_x, start_y))

                last_direction = next_direction
                if start_y == qr_size - 1:
                    # start heading down
                    if (start_x + 1, start_y) in bit_coords:
                        if (start_x + 1, start_y) not in [i[0] for i in reserve_pixels]:
                            bit_coords.append((start_x - 1, start_y))
                    orientation = 'Down'
                    dir_change = True
            elif next_direction == 'Right':
                start_x += 1
                start_y += 1
                if (start_x, start_y) in [i[0] for i in reserve_pixels]:
                    pass
                else:
                    if start_x >= 0:
                        bit_coords.append((start_x, start_y))
                last_direction = next_direction
        else:
            if next_direction == 'Left':
                start_x -= 1
                if (start_x, start_y) in [i[0] for i in reserve_pixels]:
                    pass
                else:
                    if start_x >= 0:
                        bit_coords.append((start_x, start_y))
                last_direction = next_direction
                if start_y == 0:
                    # start heading down
                    if (start_x - 1, start_y) not in [i[0] for i in reserve_pixels]:
                        bit_coords.append((start_x - 1, start_y))
                    orientation = 'Up'
                    dir_change = True
            elif next_direction == 'Right':
                start_x += 1
                start_y -= 1
                if (start_x, start_y) in [i[0] for i in reserve_pixels]:
                    pass
                else:
                    if start_x >= 0:
                        bit_coords.append((start_x, start_y))
                last_direction = next_direction

    data_pixels = []
    binary_index = 0
    for io in binary:
        line = bit_coords[binary_index]
        if io == '1':
            state_bit = 1
        else:
            state_bit = 0
        binary_index += 1
        data_pixels.append((line, state_bit))

    return data_pixels

def get_block_data(group_1_blocks, data_codewords_per_block_1, group_2_blocks, data_codewords_per_block_2, binary_values):
    block_list = []
    codeword_index = 0
    for group in range(group_1_blocks):
        codeword_list = []
        for codeword in range(data_codewords_per_block_1):
            codeword_list.append(binary_values[codeword_index])
            codeword_index += 1
        block_list.append(codeword_list)

    if group_2_blocks != '':
        for group in range(group_2_blocks):
            codeword_list = []
            for codeword in range(data_codewords_per_block_2):
                codeword_list.append(binary_values[codeword_index])
                codeword_index += 1
            block_list.append(codeword_list)

    return block_list

def equalize_list(list_of_lists):
    max_length = max(len(sublist) for sublist in list_of_lists)
    for sublist in list_of_lists:
        sublist.extend([''] * (max_length - len(sublist)))
    return list_of_lists

def interlace_lists(lists):
    interlaced = [[i[x] for i in lists] for x in range(len(lists[0]))]
    flattened_list = [item for sublist in interlaced for item in sublist]
    x = len(lists[0])
    split_list = [flattened_list[i:i + x] for i in range(0, len(flattened_list), x)]

    return split_list

def interlace_data(block_ecc_codewords, block_list):
    full_codeword_list = []
    int_blocks = [[binary_to_decimal(b) for b in i] for i in block_list]
    int_blocks = equalize_list(int_blocks)
    interlaced_codewords = interlace_lists(int_blocks)
    for sublist in interlaced_codewords:
        if '' in sublist:
            while '' in sublist:
                sublist.remove('')

    interlaced_ecc_codewords = interlace_lists(block_ecc_codewords)
    full_interlaced_codewords = [format(int(item), 'b').zfill(8) for sublist in interlaced_codewords for item in sublist]
    full_interlaced_ecc_codewords = [format(int(item), 'b').zfill(8) for sublist in interlaced_ecc_codewords for item in sublist]
    [full_codeword_list.append(i) for i in full_interlaced_codewords]
    [full_codeword_list.append(i) for i in full_interlaced_ecc_codewords]
    return full_codeword_list

def get_remainder_bits(version):
    # remainder bits are pre-defined and dependent on the qr version
    remainder_bits = [
        (1, 0),
        (2, 7),
        (3, 7),
        (4, 7),
        (5, 7),
        (6, 7),
        (7, 0),
        (8, 0),
        (9, 0),
        (10, 0),
        (11, 0),
        (12, 0),
        (13, 0),
        (14, 3),
        (15, 3),
        (16, 3),
        (17, 3),
        (18, 3),
        (19, 3),
        (20, 3),
        (21, 4),
        (22, 4),
        (23, 4),
        (24, 4),
        (25, 4),
        (26, 4),
        (27, 4),
        (28, 3),
        (29, 3),
        (30, 3),
        (31, 3),
        (32, 3),
        (33, 3),
        (34, 3),
        (35, 0),
        (36, 0),
        (37, 0),
        (38, 0),
        (39, 0),
        (40, 0),
    ]
    return [i[1] for i in remainder_bits if i[0] == int(version)][0]

def analyze_data(input_text, ecc_level):
    Numeric = False
    Byte = False
    Alphanumeric = True
    ecc_type = ''
    binary_values = []

    try:
        numbers = [int(i) for i in input_text]
        Numeric = True
        Alphanumeric = False
        ecc_type = 'Numeric'
        cc_index = 1
    except:
        for i in input_text:
            if i in alpha_numeric['Chars']:
                Alphanumeric = True
                ecc_type = 'Alphanumeric'
                cc_index = 2
            else:
                Byte = True
                Alphanumeric = False
                ecc_type = 'Byte'
                cc_index = 3
                break

    if Numeric:
        x = 3
        input_text_list = [input_text[y - x:y] for y in range(x, len(input_text) + x, x)]

        code_words = len(input_text_list)

        for i in input_text_list:
            if len(str(i)) == 2:
                binary = format(int(i), 'b').zfill(7)
            elif len(str(i)) == 1:
                binary = format(int(i), 'b').zfill(4)
            else:
                binary = format(int(i), 'b').zfill(10)
            binary_values.append(binary)

    if Alphanumeric:
        # split string into segments of length 2
        x = 2
        input_text_list = [input_text[y - x:y] for y in range(x, len(input_text) + x, x)]

        char_values = []
        code_words = len(input_text_list)
        # determine alphanumeric binary value
        for segment in input_text_list:
            char_num = 0
            char_value = 0
            if len(segment) == 2:
                char1value = alpha_numeric[segment[0]]
                char2value = alpha_numeric[segment[1]]
                char_value = (char1value * 45) + char2value
            else:
                char_value = alpha_numeric[segment]
            char_values.append(char_value)
        # convert each value to binary

        value_index = 0
        for item in char_values:
            segment = input_text_list[value_index]
            if len(input_text) % 2 != 0:
                #if item == char_values[-1]:
                if value_index == len(char_values) - 1:
                    binary = format(int(item), 'b').zfill(6)
                else:
                    binary = format(int(item), 'b').zfill(11)
            else:
                binary = format(int(item), 'b').zfill(11)
            binary_values.append(binary)
            value_index += 1

    if Byte:
        code_words = 0
        for i in input_text:
            try:
                i.encode('utf-8')
                binary = bin(int(ord(i)))[2:].zfill(8)
            except:
                break
            code_words += 1
            binary_values.append(binary)

    mode = mode_indicators[ecc_type]
    version_capacity_data = character_capacities[ecc_level]['Value_Data']
    version = min([i[0] for i in version_capacity_data if i[cc_index] > len(input_text)])
    capacity = [i[cc_index] for i in version_capacity_data if i[0] == version][0]
    ecc_code = f'{version}-{ecc_level[0]}'
    ec_codewords_per_block = ecc_dict[ecc_code]['BlockQty'] #[i[2] for i in ecc_list if i[0] == ecc_code][0]
    char_len = '9' if version <= 9 else ('26' if version <= 26 else '40')
    character_count = format(len(input_text), 'b').zfill(char_fill[ecc_type][char_len])

    group_1_blocks = ecc_dict[ecc_code]['G1Blocks'] #[i[3] for i in ecc_list if i[0] == ecc_code][0]
    data_codewords_per_block_1 = ecc_dict[ecc_code]['B1CodeQty'] #[i[4] for i in ecc_list if i[0] == ecc_code][0]
    group_2_blocks = ecc_dict[ecc_code]['G2Blocks'] #[i[5] for i in ecc_list if i[0] == ecc_code][0]
    data_codewords_per_block_2 = ecc_dict[ecc_code]['B2CodeQty'] #[i[6] for i in ecc_list if i[0] == ecc_code][0]

    total_blocks = int(group_1_blocks) + int(group_2_blocks) if group_2_blocks != '' else group_1_blocks
    total_codewords = int(int(data_codewords_per_block_1)*int(group_1_blocks)) + (int(int(data_codewords_per_block_2)*int(group_2_blocks)) if group_2_blocks != '' else 0)

    return version, capacity, ecc_code, mode, character_count, ec_codewords_per_block, group_1_blocks, data_codewords_per_block_1, group_2_blocks, data_codewords_per_block_2, total_blocks, binary_values, total_codewords






def binary_to_string(bits):
    # Convert a list of binary strings (bits) to their corresponding ASCII characters.
    return [chr(int(i, 2)) for i in bits]

def binary_to_int(bits):
    # Convert a list of binary strings (bits) to a list of integers.
    int_list = []
    for n in bits:
        dec_value = binary_to_float(n)
        int_list.append(dec_value)  # Append the converted decimal value to the list.

    return int_list  # Return the list of integers.

def binary_to_decimal(n):
    # Function to convert binary to ASCII value
    num = str(n)  # Convert the binary input to string for processing.
    dec_value = 0  # Initialize decimal value to accumulate the result.

    # Initialize base value to 1, which represents 2^0.
    base1 = 1

    len1 = len(num)  # Get the length of the binary string.
    # Iterate over the binary string from right to left.
    for i in range(len1 - 1, -1, -1):
        if (num[i] == '1'):  # If the current bit is '1', add the current base to the decimal value.
            dec_value += base1
        base1 = base1 * 2  # Move to the next base (2^n).

    return dec_value;

def binary_to_hex(bits):
    # Convert a list of binary strings (bits) to their corresponding hexadecimal representations.
    char = [chr(int(i, 2)) for i in bits]  # Convert binary to characters.
    # Format each character's ASCII value as a two-digit hexadecimal string.
    return ["{0:02x}".format(ord(i)) for i in char]

def int_to_binary(bits):
    # Convert a list of integers to a list of 8-bit binary strings.
    binary_list = []
    for num in bits:
        # Format the integer as binary and zero-fill to ensure it is 8 bits long.
        binary_data = format(int(num), 'b').zfill(8)
        binary_list.append(binary_data)  # Append the binary string to the list.

    return binary_list  # Return the list of binary strings.

def int_to_hex(bits):
    # Convert a list of integers to their hexadecimal representation by first converting to binary.
    return binary_to_hex(int_to_binary(bits))



def make_rows(mask_data, qr_size):
    rows = qr_size
    row_start = 0
    row_end = qr_size
    row_list = []
    for i in range(rows):
        states = [i[1] for i in mask_data[row_start:row_end]]
        row_list.append(states)
        row_end += rows
        row_start += rows
    return row_list

def transpose(input_list):
    output_list = []
    for i in range(len(input_list[0])):
        row = []
        for item in input_list:
            # appending to new list with values and index positions
            # i contains index position and item contains values
            row.append(item[i])
        output_list.append(row)
    return output_list

def condition_1(mask_data, qr_size):

    def count_penalty_2(input_list):
        all_runs = []
        for row in input_list:
            last_module = 0
            module_index = 0
            current_run = []
            row_run = []
            for module in row:
                if module_index == 0:
                    current_run.append(module)
                    pass
                else:
                    if module == last_module:
                        current_run.append(module)
                    else:
                        if len(current_run) == 5:
                            row_run.append(3)
                        if len(current_run) > 5:
                            row_run.append(3)
                            row_run.append(len(current_run) - 5)
                        current_run = []
                        current_run.append(module)
                module_index += 1
                last_module = module
            all_runs.append(row_run)
        return sum([sum(i) for i in all_runs])

    horizontal = make_rows(mask_data, qr_size)
    total_h_penalty = count_penalty_2(horizontal)

    vertical = []
    col_index = 0
    for i in range(qr_size):
        col_list = []
        for row in horizontal:
            col_list.append(row[col_index])
        col_index += 1
        vertical.append(col_list)
    total_v_penalty = count_penalty_2(vertical)
    return total_v_penalty + total_h_penalty

def condition_2(mask_data, qr_size):
    rows = qr_size
    row_start = 0
    row_end = qr_size
    array = []
    array_2 = []
    for i in range(rows):
        states = [i[1] for i in mask_data[row_start:row_end]]
        array.append(states)
        states_2 = []
        for state in states:
            if state == 1:
                states_2.append(0)
            else:
                states_2.append(1)
        row_end += rows
        row_start += rows
        array_2.append(states_2)

    def square_check(array):
        count = 0
        for i in range(len(array) - 1):
            for j in range(len(array[i]) - 1):
                if array[i][j] == array[i][j+1] == array[i+1][j] == array[i+1][j+1] == 1:
                    count += 1
        return count

    total_black_squares = square_check(array)
    total_white_squares = square_check(array_2)
    return (total_black_squares + total_white_squares) * 3

def condition_3(mask_data, qr_size):
    horizontals = make_rows(mask_data, qr_size)
    horizontals = transpose(horizontals)

    def count_pattern_occurrences(pattern, sequence):
        pattern_length = len(pattern)
        count = 0
        for i in range(len(sequence) - pattern_length + 1):
            if sequence[i:i + pattern_length] == pattern:
                count += 1
        return count

    pattern_occurances = 0
    for row in horizontals:
        if count_pattern_occurrences([1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0], row) != 0:
            pattern_occurances += 40
        if count_pattern_occurrences([0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1], row) != 0:
            pattern_occurances += 40

    vertical = []
    col_index = 0
    horizontals = make_rows(mask_data, qr_size)
    for i in range(qr_size):
        col_list = []
        for row in horizontals:
            col_list.append(row[col_index])
        col_index += 1
        vertical.append(col_list)

    for row in vertical:
        [row.append(0) for i in range(4)]
        [row.insert(0, 0) for i in range(4)]

    for row in vertical:
        if count_pattern_occurrences([1, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0], row) != 0:
            pattern_occurances += 40
        if count_pattern_occurrences([0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 1], row) != 0:
            pattern_occurances += 40

    return pattern_occurances

def condition_4(mask_data):
    total_on = [i[1] for i in mask_data if i[1] == 1]

    ratio = (len(total_on) / len(mask_data)) * 100
    prev_multiple = (ratio // 5) * 5
    next_multiple = ((ratio // 5) * 5) + 5

    abs_1 = int(abs(prev_multiple - 50) / 5)
    abs_2 = int(abs(next_multiple - 50) / 5)
    penalty = min([abs_1, abs_2]) * 10
    return penalty

def get_penalty(mask_data, reserve_data, qr_size):
    total_data = mask_data
    [total_data.append(i) for i in reserve_data]

    sorted_list = sorted(total_data, key=lambda x: (x[0][1], x[0][0]))

    total_penalty_1 = condition_1(sorted_list, qr_size)
    total_penalty_2 = condition_2(sorted_list, qr_size)
    total_penalty_3 = condition_3(sorted_list, qr_size)
    total_penalty_4 = condition_4(sorted_list)
    total_penalty = total_penalty_1 + total_penalty_2 + total_penalty_3 + total_penalty_4
    return total_penalty



import time
import os

def write_qr(qr_data, module_size, file_dir):
    file_start = '''
%!PS-Adobe-3.0 EPSF-3.0
%%Creator: PyQR
%%Title: QR.eps
%%CreationDate: 20240610
%%Pages: 1
%%BoundingBox: 0 0 612 792
%%Document-Fonts: Helvetica
%%LanguageLevel: 3
%%EndComments
%%BeginProlog
%%EndProlog
%%Page: 1 1
save'''
    add_module = f'''
/module_size {module_size} def\n'''
    add_bits = '''
/on_bit {gsave 0 0 0 1 setcmykcolor newpath 0 0 moveto 0 module_size lineto module_size module_size lineto module_size 0 lineto 0 0 lineto fill stroke grestore } bind def
/off_bit {gsave 0 0 0 0 setcmykcolor newpath 0 0 moveto 0 module_size lineto module_size module_size lineto module_size 0 lineto 0 0 lineto fill stroke grestore } bind def
/qr_code {
    gsave
'''

    file_end = '''
    grestore
} bind def
20 20 translate qr_code
restore showpage
%
% End of page
%
%%Trailer
%%EOF'''


    if file_dir is None:
        filename = f'qr_{time.time()}.ps'
        qr_file = open(filename, 'x')
    else:
        try:
            filename = file_dir
            qr_file = open(filename, 'x')
        except:
            raise ValueError(f"{file_dir} Does Not Exist")

    qr_file.write(file_start)
    qr_file.write(add_module)
    qr_file.write(add_bits)
    state = 'off_bit'
    for coord, state in qr_data:
        x, y = coord
        if state == 1:
            state = 'on_bit'
        else:
            state = 'off_bit'
        qr_file.write(f'''
    gsave {x*module_size} {y*module_size} translate {state} grestore''')
    qr_file.write(file_end)
    qr_file.close()

    os.startfile(filename)

def get_ps_string(qr_data, module_size):
    full_ps_string = ''
    file_start = '''
/qr_code {
    gsave\n'''
    add_module = f'''
    /module_size {module_size} def\n'''
    add_bits = '''
    /on_bit {gsave 0 0 0 1 setcmykcolor newpath 0 0 moveto 0 module_size lineto module_size module_size lineto module_size 0 lineto 0 0 lineto fill stroke grestore } bind def
    /off_bit {gsave 0 0 0 0 setcmykcolor newpath 0 0 moveto 0 module_size lineto module_size module_size lineto module_size 0 lineto 0 0 lineto fill stroke grestore } bind def
    /qr_data {
        gsave
    '''

    file_end = '''
        grestore
    } bind def
    20 20 translate qr_data
    grestore
} bind def
'''

    full_ps_string += file_start
    full_ps_string += add_module
    full_ps_string += add_bits

    state = 'off_bit'
    for coord, state in qr_data:
        x, y = coord
        if state == 1:
            state = 'on_bit'
        else:
            state = 'off_bit'
        full_ps_string += f'''
        gsave {x*module_size} {y*module_size} translate {state} grestore'''

    full_ps_string += file_end

    return full_ps_string



from setuptools import setup, find_packages

setup(
    name='PiQR',
    version='0.1.5',
    author='PimpDiCaprio',
    author_email='info@aperturesoftware.us',
    description='A Python library for generating and displaying QR codes.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/PimpDiCaprio/PiQR',
    packages=find_packages(),
    scripts=['PiQR.py'],
    install_requires=[
        'opencv-contrib-python',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)


import cv2
import numpy as np
import os
import math

def correct_orientation(input_data, qr_size, module_size):
    # Corrects the orientation of the QR code coordinates.
    new_coords = []
    for coords, state in input_data:
        x = coords[0]  # X-coordinate
        y = qr_size - coords[1]  # Flip Y-coordinate to match image orientation
        new_coords.append(((x, y), state))  # Append corrected coordinates with state
    return new_coords  # Return the list of new coordinates

def write_image(image, file_dir):
    # Saves the given image to the specified directory.
    if file_dir is None:
        # If no directory is specified, save to a default filename.
        cv2.imwrite('qr_image.png', image)
    else:
        try:
            # Attempt to save the image to the specified directory.
            cv2.imwrite(file_dir, image)
        except:
            # Raise an error if the directory does not exist.
            raise ValueError(f"{file_dir} Does Not Exist")

def create_qr_png(input_data, file_dir, code_shape='Square'):
    # Generates a QR code image based on input data and saves it to the specified directory.
    qr_size = int(math.sqrt(len(input_data)))  # Calculate the size of the QR code
    module_size = 1  # Size of each module in the QR code
    size_factor = 80  # Scaling factor for module size
    placement = 100  # Placement offset for the QR code in the image
    oriented_data = correct_orientation(input_data, qr_size, module_size)  # Correct the orientation

    # Calculate the overall width and height of the QR code image
    width = (qr_size * size_factor) + (placement * 2)
    height = width
    # Create a blank white image for the QR code
    qr_image = np.zeros([width + (module_size * size_factor) + placement, height, 3], dtype=np.uint8)
    #qr_image.fill(255)  # Fill the image with white color
    cv2.rectangle(qr_image, (0, 0), (width+(module_size*size_factor)+placement, width+(module_size*size_factor)+placement), (199, 199, 191), -1)

    # Draw the QR code modules on the image
    for coords, state in oriented_data:
        x, y = coords  # Get the coordinates
        # Calculate the corners of the module
        x1 = x * size_factor + placement - 50
        y1 = y * size_factor + placement
        x2 = (x + 1) * size_factor + placement - 50
        y2 = (y + 1) * size_factor + placement

        if state != 0:
            color = (0, 0, 0)  # Set color to black for filled modules
            # Draw either a circle or rectangle based on the code_shape parameter
            if code_shape == 'Circle':
                cv2.circle(qr_image, (x1 + (x2 - x1), y1 + (y2 - y1)), 40, color, thickness=-1)
            else:
                cv2.rectangle(qr_image, (x1 + 50, y1), (x2 + 50, y2), color, -1)

    # Resize the QR code image to a standard size for output
    bigger = cv2.resize(qr_image, (120, 120))
    write_image(bigger, file_dir)  # Save the generated image

def display_qr_image(input_data, code_shape):
    # Displays the QR code image in a window.
    alignment_data = []  # Initialize an empty list for alignment data
    module_size = 1  # Size of each module in the QR code
    size_factor = 80  # Scaling factor for module size
    qr_size = int(math.sqrt(len(input_data)))  # Calculate the size of the QR code
    placement = 100  # Placement offset for the QR code in the image
    oriented_data = correct_orientation(input_data, qr_size, module_size)  # Correct the orientation
    oriented_alignment = correct_orientation(alignment_data, qr_size, module_size)  # Correct alignment orientation

    # Calculate the overall width and height of the QR code image
    width = (qr_size * size_factor) + (placement * 2)
    height = width
    # Create a blank white image for the QR code
    qr_image = np.zeros([width + (module_size * size_factor) + placement, height, 3], dtype=np.uint8)
    qr_image.fill(255)  # Fill the image with white color

    # Draw the QR code modules on the image
    for coords, state in oriented_data:
        x, y = coords  # Get the coordinates
        # Calculate the corners of the module
        x1 = x * size_factor + placement - 50
        y1 = y * size_factor + placement
        x2 = (x + 1) * size_factor + placement - 50
        y2 = (y + 1) * size_factor + placement

        if state != 0:
            # Determine color based on whether the coordinates are part of the alignment data
            if coords in [i[0] for i in oriented_alignment]:
                color = (0, 255, 0)  # Set color to green for alignment modules
            else:
                color = (0, 0, 0)  # Set color to black for filled modules

            # Draw either a circle or rectangle based on the code_shape parameter
            if code_shape == 'Circle':
                cv2.circle(qr_image, (x1 + (x2 - x1), y1 + (y2 - y1)), 40, color, thickness=-1)
            else:
                cv2.rectangle(qr_image, (x1 + 50, y1), (x2 + 50, y2), color, -1)

    # Display the generated QR code image in a window
    cv2.imshow('QR', cv2.resize(qr_image, (600, 600)))
    cv2.waitKey(0)  # Wait indefinitely until a key is pressed


import math

def mask_0(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if (row + column) % 2 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_1(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if (row) % 2 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_2(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if (column) % 3 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_3(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if (row + column) % 3 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_4(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if 	(math.floor(row / 2) + math.floor(column / 3) ) % 2 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_5(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if ((row * column) % 2) + ((row * column) % 3) == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_6(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if (((row * column) % 2) + ((row * column) % 3)) % 2 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data

def mask_7(data_list, qr_size):
    masked_data = []
    for coords, state in data_list:
        column = coords[0]
        row = (qr_size-1) - int(coords[1])
        if (((row + column) % 2) + ((row * column) % 3)) % 2 == 0:
            if state == 0:
                state = 1
            else:
                state = 0
        masked_data.append((coords, state))
    return masked_data


import math
from PiQR.qr_shapes import *
from PiQR.error_correction import *
from PiQR.qr_masks import *
from PiQR.qr_penalty import *
from PiQR.qr_imager import *
from PiQR.qr_ps import *
from PiQR.qr_coords import *
from PiQR.data_analysis import *

# Resource Links i used to build this
# full qr creation Tutorial
# https://www.thonky.com/qr-code-tutorial/introduction
# General Resource QR Wiki
# https://en.wikipedia.org/wiki/QR_code
# QR Creation Tool with Steps
# https://www.nayuki.io/page/creating-a-qr-code-step-by-step


def generate_qr(data_string, correction_level='Low', mask=None, verbose=False):
    # Correction Level Options: Low, Medium, Quartile, High - Defaulted to Low
    if mask is not None:
        if mask not in [0, 1, 2, 3, 4, 5, 6, 7]:
            raise Exception('Invalid Mask Number - Please Select a Number Between 0 - 7')
            return

    if correction_level not in ['Low', 'Medium', 'Quartile', 'High']:
        raise Exception('Invalid Correction Level - Please Select from (Low, Medium, Quartile, High)')
        return

    data_string = str(data_string)
    
    (version, capacity, ecc_code, mode, character_count, ec_codewords_per_block, group_1_blocks,
     data_codewords_per_block_1, group_2_blocks, data_codewords_per_block_2, total_blocks, binary_values,
     total_codewords) = analyze_data(data_string, correction_level)
    print(f'Total Group 1 Blocks: {group_1_blocks}') if verbose else ''
    print(f'Total Blocks per Group 1 Blocks: {data_codewords_per_block_1}') if verbose else ''
    print(f'Total Group 2 Blocks: {group_2_blocks}') if verbose else ''
    print(f'Total Blocks per Group 2 Blocks: {data_codewords_per_block_2}') if verbose else ''

    total_data_bits = total_codewords * 8
    base_message = "".join(binary_values)
    print(f'Total Data Bits Required: {total_data_bits}') if verbose else ''
    print(f'Requird Codewords Per Block: {ec_codewords_per_block}') if verbose else ''
    print(f'Version: {version}') if verbose else ''
    print(f'Mode String: {mode}') if verbose else ''
    print(f'Character String: {character_count}') if verbose else ''
    print(f'Initial Data String: {base_message}') if verbose else ''

    binary = ''
    binary += mode
    binary += character_count
    binary += base_message

    terminator = '0000' if len(binary) + 4 < total_data_bits else (
        '0'.zfill(total_data_bits - len(binary)) if len(binary) < total_data_bits else '')
    print(f'Terminator String: {terminator}') if verbose else ''

    binary += terminator

    while len(binary) % 8 != 0:  # the length has to be a multiple of 8
        binary += '0'

    pad_bits = int((total_data_bits - len(binary)) / 8)
    print(f'Total Pad Bits Required: {pad_bits}') if verbose else ''
    pad_bit_1 = '11101100'  # qr spec pad bit #1
    pad_bit_2 = '00010001'  # qr spec pad bit #2
    pad_string = ''
    while pad_bits > 0:
        binary = binary + pad_bit_1
        pad_string = pad_string + pad_bit_1
        pad_bits -= 1
        if pad_bits == 0:
            pass
        else:
            binary = binary + pad_bit_2
            pad_string = pad_string + pad_bit_2
            pad_bits -= 1
    print(f'Pad String: {pad_string}') if verbose else ''
    print(f'Data String: {binary}') if verbose else ''
    if total_blocks > 1:
        # split binary string into segments of length 8
        x = 8
        input_text_list = [binary[y - x:y] for y in range(x, len(binary) + x, x)]
        block_ecc_codewords = []
        block_codewords = get_block_data(group_1_blocks, data_codewords_per_block_1, group_2_blocks,
                                                   data_codewords_per_block_2, input_text_list)
        for block in block_codewords:
            block_ecc_codewords.append(get_ecc(ec_codewords_per_block, "".join(block)))

        interlaced_binary_data = interlace_data(block_ecc_codewords, block_codewords)
        binary = "".join(["".join(i) for i in interlaced_binary_data])
    else:

        ecc_code_words = get_ecc(ec_codewords_per_block, binary)
        binary += "".join([format(int(item), 'b').zfill(8) for item in ecc_code_words])

    remainder_bits = ''.zfill(get_remainder_bits(version))
    binary += remainder_bits
    print(f'Full Binary Data: {binary}') if verbose else ''

    reserve_pixels = []
    module_size = 1
    qr_size = (((int(version) - 1) * 4) + 21)
    print(f'QR Size: {qr_size} x {qr_size}') if verbose else ''
    finder_dim = 8


    # first set the grid locations of the first 3 reserve patters
    reserve1x, reserve1y = finder_dim, 0
    [reserve_pixels.append(i) for i in get_coords(reserve1x, reserve1y, get_reserve_1_shape())]

    reserve2x, reserve2y = 0, (qr_size - (finder_dim * 2)) + finder_dim - 1
    [reserve_pixels.append(i) for i in get_coords(reserve2x, reserve2y, get_reserve_2_shape())]

    reserve3x, reserve3y = (qr_size - (finder_dim * 2)) + finder_dim, (qr_size - (finder_dim * 2)) + finder_dim - 1
    [reserve_pixels.append(i) for i in get_coords(reserve3x, reserve3y, get_reserve_3_shape())]


    # now the three finders can be placed
    finder1x, finder1y = 0, qr_size - finder_dim  # the top left finder
    [reserve_pixels.append(i) for i in get_coords(finder1x, finder1y, get_finder_1_shape()) if
     i[0] not in [i[0] for i in reserve_pixels]]

    finder2x, finder2y = qr_size - finder_dim, qr_size - finder_dim  # the top right finder
    [reserve_pixels.append(i) for i in get_coords(finder2x, finder2y, get_finder_2_shape()) if
     i[0] not in [i[0] for i in reserve_pixels]]

    finder3x, finder3y = 0, 0  # the bottom left finder
    [reserve_pixels.append(i) for i in get_coords(finder3x, finder3y, get_finder_3_shape()) if
     i[0] not in [i[0] for i in reserve_pixels]]


    v_reserver_dim = 3
    v_reserve1x, v_reserve1y = ((qr_size - (finder_dim * 2)) + finder_dim) - v_reserver_dim, (
                (qr_size - (finder_dim * 2)) + finder_dim) + 2
    v_reserve2x, v_reserve2y = 0, finder_dim
    if version >= 7:
        [reserve_pixels.append(i) for i in get_coords(v_reserve1x, v_reserve1y, get_v_reserve_1_shape()) if
         i[0] not in [i[0] for i in reserve_pixels]]
        [reserve_pixels.append(i) for i in get_coords(v_reserve2x, v_reserve2y, get_v_reserve_2_shape()) if
         i[0] not in [i[0] for i in reserve_pixels]]

    # Mark the dark module location
    dark_module_coord = ((4 * version) + 9, finder_dim)
    dark_x, dark_y = (qr_size) - ((dark_module_coord[0])), (dark_module_coord[1]) - 1
    reserve_pixels.append(((dark_x, dark_y), 1))

    # alignment pattern only added if the version is greater than 1
    if version > 1:
        alignment_patterns = get_alignment_placements(version, reserve_pixels, qr_size)

    reserve_pixels = get_timing_coords(qr_size, reserve_pixels)

    data_pixels = place_data_codewords(qr_size, binary, reserve_pixels)

    masks = [
        (0, mask_0),
        (1, mask_1),
        (2, mask_2),
        (3, mask_3),
        (4, mask_4),
        (5, mask_5),
        (6, mask_6),
        (7, mask_7)
    ]

    if mask is None:
        penalty_list = []
        for mask_int, mask_function in masks:
            masked_data = mask_function(data_pixels, qr_size)
            mask_penalty = get_penalty(masked_data, reserve_pixels, qr_size)
            penalty_list.append((masked_data, mask_penalty, mask_int))

        min_penalty = min([i[1] for i in penalty_list])
        print(f'Selected Mask: {[x[2] for x in penalty_list if x[1] == min_penalty][0]}') if verbose else ''
        mask_number = format([x[2] for x in penalty_list if x[1] == min_penalty][0], 'b').zfill(3)
        qr_data = [x[0] for x in penalty_list if x[1] == min_penalty][0]
    else:
        selected_mask = [i[1] for i in masks if i[0] == mask][0]
        print(f'Selected Mask: {mask}') if verbose else ''
        qr_data = selected_mask(data_pixels, qr_size)
        [qr_data.append(i) for i in reserve_pixels]
        mask_number = format(mask, 'b').zfill(3)

    # Calculate the binary for the format version string or use the list in the ECC script to get it
    error_correction_bits = [('Low', '01', 1), ('Medium', '00', 0), ('Quartile', '11', 3), ('High', '10', 2)]
    format_version = [i[1] for i in error_correction_bits if i[0] == correction_level][0] + mask_number

    ecc_bits = get_format_ecc(10, format_version)
    ecc_format_version = format_version + ecc_bits

    set_xor_string = '101010000010010'  # qr spec xor string for all qr types
    final_string = ''
    bit_index = 0
    for bit in ecc_format_version:
        xor_result = int(bit) ^ int(set_xor_string[bit_index])
        bit_index += 1
        final_string = final_string + str(xor_result)

    updated_reserve = []

    reserve_1_shape = get_reserve_1_shape(binary=final_string[0:6])
    reserve_2_shape = get_reserve_2_shape(binary=final_string)
    reserve_3_shape = get_reserve_3_shape(binary=final_string[7:])

    [updated_reserve.append(i) for i in get_coords(reserve1x, reserve1y, reserve_1_shape)]
    [updated_reserve.append(i) for i in get_coords(reserve2x, reserve2y, reserve_2_shape)]
    [updated_reserve.append(i) for i in get_coords(reserve3x, reserve3y, reserve_3_shape)]

    # generate a version error correction if the version is greater than 7
    if version >= 7:
        version_string = format(int(version), 'b').zfill(6)
        ecc_bits = get_version_ecc(12, version_string)
        ecc_version_string = version_string.zfill(6) + ecc_bits
        r1 = [2, 1, 0,
              5, 4, 3,
              8, 7, 6,
              11, 10, 9,
              14, 13, 12,
              17, 16, 15]

        v_reserve_1_shape = get_v_reserve_1_shape(binary=[ecc_version_string[i] for i in r1])
        r2 = [17, 14, 11, 8, 5, 2,
              16, 13, 10, 7, 4, 1,
              15, 12, 9, 6, 3, 0]

        v_reserve_2_shape = get_v_reserve_2_shape(binary=[ecc_version_string[i] for i in r2])

        [updated_reserve.append(i) for i in get_coords(v_reserve1x, v_reserve1y, v_reserve_1_shape)]
        [updated_reserve.append(i) for i in get_coords(v_reserve2x, v_reserve2y, v_reserve_2_shape)]

    # update to current coords list with new data
    for coord, state in updated_reserve:
        reserve_coords = [i[0] for i in qr_data]
        if coord in reserve_coords:
            data_index = reserve_coords.index(coord)
            qr_data[data_index] = (coord, state)

    return [qr_data, module_size]


def write_ps(data, file_dir=None):
    qr_data, module_size = data
    write_qr(qr_data, module_size, file_dir)

def return_ps(data, mod_size=False):
    qr_data, module_size = data
    if not mod_size:
        pass
    else:
        module_size = mod_size
    ps_data = get_ps_string(qr_data, module_size)
    return ps_data

def make_png(data, file_dir, code_shape='Square'):
    qr_data, module_size = data
    create_qr_png(qr_data, file_dir, code_shape)

def show_png(data, code_shape='Square'):
    qr_data, module_size = data
    display_qr_image(qr_data, code_shape)
